/*  Network Interface Connectivity Emulator v1.0 4/26/2013
    Copyright (C) 2013  Bobby Black (Auburn University),
    Tao Lin (taolin@vt.edu), Scott F. Midkiff (Virginia Tech), and
    Shiwen Mao (Auburn University) 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <fstream>
#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <iterator>
#include <algorithm>
#include <ctime>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

// prototypes
void instantiateVector(char* argv[], vector<string> &vec);
void printVector(vector<string> vec);
void resettrafficcontrol();
void loopVector(vector<string> vec);
void parseVector(string s);

int main(int argc, char* argv[]) {
	cout << "Network Interface Connectivity Emulator Copyright (C) 2013  Bobby Black\n";
	cout << "Starting...\n";
	if (argc < 2) {
		cout << "You must specify a configuration file. Ex: ./nice example.cfg\nExiting...\n";	
		exit (EXIT_FAILURE);
	}
	vector<string> list;	// initiate an empty vector of type string
	instantiateVector(argv, list);	// read in file to vector
	printVector(list);	// print vector contents
	resettrafficcontrol();	// removes old traffic controls
	loopVector(list);	// start vector loop
	return 0;
}
void instantiateVector(char* argv[], vector<string> &vec) {
	cout << "Opening " << argv[1] << "...\n";
	ifstream ifile(argv[1]);	// open file specified in command line
	if (ifile) {
		cout << "Reading " << argv[1] << "...\n";
		string line;	// initiate an empty string
		while (getline(ifile, line))	// get every line from the file 1 by 1
			vec.push_back(line);	// save line to vector
	}
	else {
		cout << argv[1] << " was not found.\nExiting...\n";
		exit (EXIT_FAILURE);
	}
}

void printVector(vector<string> vec) {
	cout << "Printing vector...\n";
	for (unsigned i=0; i<vec.size(); i++)	// iterate through the vector
		cout << vec.at(i) << "\n";	// print vector contents at i
}

void resettrafficcontrol() {
	for (int i=0; i<=100; i++) {
		stringstream cmd,cmd2;
		//cout << "Resetting traffic controls from interfaces eth" << i << "\n";
		cmd << "tc qdisc del dev eth" << i << " root 2>/dev/null";	// deletes interface device from traffic control
		cmd2 << "tc qdisc add dev eth" << i << " root handle 1: htb default 1 root 2>/dev/null";	// adds interface device to traffic control
		system(cmd.str().c_str());
		system(cmd2.str().c_str());
	}
}

void loopVector(vector<string> vec) {
	time_t initialTime = time(0);	// declare the time as of now
	unsigned lineNumber = 1;	// start reading the configuration file from this line
	int priority = 1;	// as time passes and commands are executed, priority is used to update filters
	while (lineNumber<vec.size()) {
		string s = vec.at(lineNumber);	// put line from list into string s
		istringstream ss(s);	// convert string s to istringstream?
		if (!ss.eof()) {
			string seconds;	// here's a nice, empty string
			getline(ss, seconds, '\t');	// find a tab, save contents to seconds
			if ((time(0) - initialTime) >= atoi(seconds.c_str())) {	// if specific time has passed, start executing command
				string interface, srcip, srcmac1, srcmac2, iphandle, loss, ber, delay, delayvariation;	// declare contents of line
				getline(ss, interface, '\t');	// gets the interface from lineNumber
				getline(ss, srcip, '\t');	// gets the source ip from lineNumber
				istringstream srcipss(srcip);
				while (!srcipss.eof()) {
					getline(srcipss, iphandle, '.');	// gets the last digits of the source ip, used later for an alias
				}
				getline(ss, loss, '\t');	// gets the packet loss % from lineNumber
				//getline(ss, ber, '\t');	// gets the bit error rate from lineNumber
				getline(ss, delay, '\t');	// gets the delay from lineNumber
				getline(ss, delayvariation, '\t');	// gets the delay variation from lineNumber

				// Use source MAC alias if source IP is used.
				//if (srcip == "192.168.1.1") { srcmac1 = "c860002e"; srcmac2 = "1b7b"; } // MAC = c8:60:00:2e:1b:7b
				//else if (srcip == "192.168.1.2") { srcmac1 = "c860002e"; srcmac2 = "a699"; } // MAC = c8:60:00:2e:a6:99
				//else if (srcip == "192.168.1.3") { srcmac1 = "c860002e"; srcmac2 = "1aa9"; } // MAC = c8:60:00:2e:1a:a9

				cout << "#Time\tinterface\tsource-ip->MAC\tloss-rate\tdelay\tdelay-variation\n";	// prints a description
				if (srcmac1 != "") {
					cout << seconds << "\t" << interface << "\t" << srcip << "->" << srcmac1 << srcmac2 << "\t" << loss << "\t" << delay << "\t" << delayvariation << "\n"; // prints what is about to be the new network topology for a given interface
				}
				else {
					cout << seconds << "\t" << interface << "\t" << srcip << "\t" << loss << "\t" << delay << "\t" << delayvariation << "\n"; // prints what is about to be the new network topology for a given interface
				}
				lineNumber++;
				stringstream cmd,cmd2,cmd3,cmd4;	// system call command placeholders

				cmd << "tc class add dev " << interface << " parent 1: classid 1:" << iphandle << " htb rate 100mbit 2>/dev/null";	// creates a class for interface
				cmd2 << "tc qdisc add dev " << interface << " parent 1:" << iphandle << " handle " << iphandle << ": netem loss " << loss << "  delay " << delay << " " << delayvariation << " 2>/dev/null";	// adds a qdisc with connection attributes
				cmd3 << "tc qdisc change dev " << interface << " parent 1:" << iphandle << " handle " << iphandle << ": netem loss " << loss << " delay " << delay << " " << delayvariation << " 2>/dev/null";	// updates a qdisc with connection attributes if it already exists
				if (srcmac1 != "") {
					cmd4 << "tc filter add dev " << interface << " parent 1: protocol ip prio " << priority << " u32 match u16 0x0800 0xFFFF at -2 match u32 0x" << srcmac1 << " 0xFFFFFFFF at -8 match u16 0x" << srcmac2 << " 0xFFFF at -4 flowid 1:" << iphandle; // creates a filter for a specific source MAC, priority overrides previous filters with the same source MAC
				}
				else {
					cmd4 << "tc filter add dev " << interface << " parent 1: protocol ip prio " << priority << " u32 match ip src " << srcip << " flowid 1:" << iphandle << " 2>/dev/null";	// creates a filter for specific source IPs, priority overrides previous filters with the same source IP
				}

				system(cmd.str().c_str());	// executes cmd on the system
				system(cmd2.str().c_str());	// executes cmd2 on the system
				system(cmd3.str().c_str());	// executes cmd3 on the system
				system(cmd4.str().c_str());	// executes cmd4 on the system
				priority++;
			}
		}
		usleep(1000);	// sleeps for 1000 microseconds
	}
}

void loopParseVector(vector<string> vec) {
	for (unsigned i=0; i<vec.size(); i++) {
		string s = vec.at(i);	// put line from list into string s
		istringstream ss(s);	// convert string s to istringstream?
		while (!ss.eof())
		{
		  string x;	// here's a nice, empty string
		  getline(ss, x, '\t');	// find a tab and put that field into x
		  cout << x << "TAB";	// print it out, even if we already hit EOF
		}
		cout << "\n";	// print new line
	}
}
